/*
 * Copyright (C) 2019 Purism SPC
 * SPDX-License-Identifier: LGPL-2.1
 * Author: Julian Sparber <julian.sparber@puri.sm>
 * 
 * This file was copied from the telepathy backend and then modified
 */

using GLib;
using Gee;
using Folks;
using Folks.Backends.Purple;

extern const string BACKEND_NAME;

[DBus (name = "im.pidgin.purple.PurpleInterface")]
public interface PurpleInterface : Object {
  public signal void account_added (int account);
  public signal void account_removed (int account);
  public signal void buddy_added (int buddy);
  public signal void buddy_removed (int buddy);
  public abstract int[] purple_accounts_get_all_active () throws GLib.Error;
  public abstract string purple_account_get_protocol_id (int account) throws GLib.Error;
  public abstract string purple_account_get_username (int account) throws GLib.Error;
  public abstract string purple_account_get_alias (int account) throws GLib.Error;
  public abstract string purple_account_get_name_for_display (int account) throws GLib.Error;
  public abstract string purple_account_get_protocol_name (int account) throws GLib.Error;
  public abstract int[] purple_blist_get_buddies () throws GLib.Error;
  public abstract string purple_buddy_get_alias (int buddy) throws GLib.Error;
  public abstract string purple_buddy_get_name (int buddy) throws GLib.Error;
  public abstract int purple_buddy_get_account (int account) throws GLib.Error;
  public abstract int purple_buddy_get_icon (int buddy) throws GLib.Error;
  public abstract int purple_buddy_get_presence (int buddy) throws GLib.Error;
  public abstract string purple_buddy_icon_get_full_path (int buddy) throws GLib.Error;
}

/**
 * A backend which connects to libpurple creates a {@link PersonaStore}.
 */
public class Folks.Backends.Purple.Backend : Folks.Backend
{
  private bool _is_prepared = false;
  private bool _prepare_pending = false; /* used by unprepare() too */
  private bool _is_quiescent = false;
  private uint purple_watcher;

  private ulong _account_added_id;
  private ulong _account_removed_id;
  private ulong _buddy_added_id;
  private ulong _buddy_removed_id;

  private int[] _accounts;
  private HashMap<string, PersonaStore> _persona_stores;
  private Map<string, PersonaStore> _persona_stores_ro;
  private PurpleInterface _purple = null;


  /**
   * {@inheritDoc}
   */
  public override string name { get { return BACKEND_NAME; } }

  /**
   * {@inheritDoc}
   */

  public override Map<string, Folks.PersonaStore> persona_stores
  {
    get { return this._persona_stores_ro; }
  }

  /**
   * {@inheritDoc}
   */
  public override void enable_persona_store (Folks.PersonaStore store)
  {
    var _store = store as Purple.PersonaStore;
    if (_store == null)
      return;

    if (this.persona_stores.has_key (_store.id) == false)
      this._add_persona_store (_store);
  }

  /**
   * {@inheritDoc}
   */
  public override void disable_persona_store (Folks.PersonaStore store)
  {
    var _store = store as Purple.PersonaStore;
    if (_store == null)
      return;

    if (this.persona_stores.has_key (_store.id))
      this._remove_persona_store (_store);
  }

  /**
   * {@inheritDoc}
   */
  public override void set_persona_stores (Set<string>? storeids)
  {
    bool added_stores = false;
    PersonaStore[] removed_stores = {};

    /* First handle adding any missing persona stores. */
    foreach (string id in storeids)
    {
      if (this._persona_stores.has_key (id) == false)
      {
        try {
          PersonaStore store = new Purple.PersonaStore (int.parse (id), this._purple);
          this._add_persona_store (store, false);
          added_stores = true;
        } catch (Error e) {
          error ("Error talking to libpurple: %s", e.message);
        }
      }
    }

    foreach (PersonaStore store in this._persona_stores.values)
    {
      if (!storeids.contains (store.id))
      {
        removed_stores += store;
      }
    }

    for (int i = 0; i < removed_stores.length; ++i)
    {
      this._remove_persona_store ((PersonaStore) removed_stores[i], false);
    }

    /* Finally, if anything changed, emit the persona-stores notification. */
    if (added_stores || removed_stores.length > 0)
    {
      this.notify_property ("persona-stores");
    }
  }

  /**
   * {@inheritDoc}
   */
  public Backend ()
  {
    Object ();
  }

  construct {
    this._persona_stores = new HashMap<string, PersonaStore> ();
    this._persona_stores_ro = this._persona_stores.read_only_view;
  }

  /**
   * Whether this Backend has been prepared.
   *
   * See {@link Folks.Backend.is_prepared}.
   *
   * @since 0.3.0
   */
  public override bool is_prepared
  {
    get { return this._is_prepared; }
  }

  /**
   * Whether this Backend has reached a quiescent state.
   *
   * See {@link Folks.Backend.is_quiescent}.
   *
   * @since 0.6.2
   */
  public override bool is_quiescent
  {
    get { return this._is_quiescent; }
  }

  /**
   * {@inheritDoc}
   */
  public override async void prepare () throws GLib.Error
  {
    Internal.profiling_start ("preparing Purple.Backend");

    if (this._is_prepared || this._prepare_pending)
    {
      return;
    }

    this.purple_watcher = Bus.watch_name (BusType.SESSION,
                                          "im.pidgin.purple.PurpleService",
                                          BusNameWatcherFlags.NONE,
                                          _purple_connect,
                                          _purple_disconnect);

    Internal.profiling_end ("preparing Purple.Backend");
  }

  /**
   * {@inheritDoc}
   */
  public override async void unprepare () throws GLib.Error
  {
    Bus.unwatch_name (this.purple_watcher);


    _purple_disconnect ();
  }

  private void _purple_connect () {
    try
    {
      this._prepare_pending = true;
      this.freeze_notify ();

      this._purple = Bus.get_proxy_sync (BusType.SESSION,
                                         "im.pidgin.purple.PurpleService",
                                         "/im/pidgin/purple/PurpleObject");

      this._accounts = this._purple.purple_accounts_get_all_active ();
      foreach (int account in this._accounts) {
        string protocol = this._purple.purple_account_get_protocol_id (account);
        // Create PersonaStore for each account but sms
        if (protocol == "prpl-mm-sms")
          continue;

        PersonaStore store = new Purple.PersonaStore (account, this._purple);
        this._add_persona_store (store);
      }

      this._account_added_id = this._purple.account_added.connect ( (account) => {
        if (this._persona_stores.has_key (account.to_string ()))
          return;
        try {

          string protocol = this._purple.purple_account_get_protocol_id (account);
          if (protocol == "prpl-mm-sms")
            return;

          PersonaStore store = new Purple.PersonaStore (account, this._purple);
          this._add_persona_store (store);
          debug ("Account added");
        } catch (Error e) {
          error ("Error talking to libpurple: %s", e.message);
        }
      });

      this._account_removed_id = this._purple.account_removed.connect ( (account) => {
        var store = this._persona_stores.get (account.to_string ());
        if (store == null)
          return;

        this._remove_persona_store (store);
        debug ("Account removed");
      });

      this._buddy_added_id = this._purple.buddy_added.connect ( (buddy) => {
        try {
          int account = this._purple.purple_buddy_get_account (buddy);
          string protocol = this._purple.purple_account_get_protocol_id (account);
          if (protocol == "prpl-mm-sms")
            return;

          var store = this._persona_stores.get (account.to_string ()) as Purple.PersonaStore;
          if (store != null) {
            store._add_persona (buddy);
          }
        } catch (Error e) {
          error ("Error talking to libpurple: %s", e.message);
        }
      });

      this._buddy_removed_id = this._purple.buddy_removed.connect ( (buddy) => {
        /* We can't ask purple for the account of the persona store since purple already removed it */
        foreach (var store in this._persona_stores.values)
          store._remove_persona (buddy);
      });


      this._is_prepared = true;
      this.notify_property ("is-prepared");

      this._is_quiescent = true;
      this.notify_property ("is-quiescent");
    }
    catch (Error e) {
      error ("Error talking to libpurple: %s", e.message);
    }
    finally
    {
      this.thaw_notify ();
      this._prepare_pending = false;
    }
  }

  private void _purple_disconnect () {
    if (!this._is_prepared || this._prepare_pending)
    {
      return;
    }

    try
    {
      this._prepare_pending = true;
      this.freeze_notify ();

      if (this._purple != null) {
       this._purple.disconnect (this._account_added_id);
       this._purple.disconnect (this._account_removed_id);
       this._purple = null;

       var iter = this._persona_stores.map_iterator ();
       while (iter.next () == true)
       {
         this._persona_store_removed_cb (iter.get_value () as Folks.PersonaStore);
         iter.unset ();
       }
      }

      this._is_quiescent = false;
      this.notify_property ("is-quiescent");

      this._is_prepared = false;
      this.notify_property ("is-prepared");
    }
    finally
    {
      this.thaw_notify ();
      this._prepare_pending = false;
    }
  }

  private void _add_persona_store (PersonaStore store, bool notify = true)
  {
    this._persona_stores.set (store.id, store);

    store.removed.connect (this._persona_store_removed_cb);
    this.persona_store_added (store);
    debug ("Purple persona store was added");
    if (notify)
      this.notify_property ("persona-stores");
  }

  private void _remove_persona_store (PersonaStore store, bool notify = true)
  {
    this._persona_stores.unset (store.id);
    store.removed.disconnect (this._persona_store_removed_cb);
    this.persona_store_removed (store);

    if (notify)
      this.notify_property ("persona-stores");
  }

  private void _persona_store_removed_cb (Folks.PersonaStore store)
  {
    store.removed.disconnect (this._persona_store_removed_cb);
    this.persona_store_removed (store);
    this.notify_property ("persona-stores");
  }
}
