/*
 * Copyright (C) 2019 Purism SPC
 * SPDX-License-Identifier: LGPL-2.1
 * Author: Julian Sparber <julian.sparber@puri.sm>
 * 
 * This file was copied from the telepathy backend and then modified
 */

using GLib;
using Gee;
using GLib;
using Gee;
using Folks;

extern const string G_LOG_DOMAIN;

/**
 * A persona store which is associated with a single Purple account. It will
 * create {@link Persona}s for each of the contacts in the account's
 * contact list.
 *
 */
public class Folks.Backends.Purple.PersonaStore : Folks.PersonaStore
{
  private PurpleInterface _purple = null;
  public string protocol_name { get; set; }
  private int account { get; set; }

  private string[] _always_writeable_properties = {};

  private HashMap<string, Persona> _personas;
  private Map<string, Persona> _personas_ro;

  /* Buddy to Persona map */
  private HashMap<int, Persona> _buddy_persona_map;

  private bool _is_prepared = false;
  private bool _prepare_pending = false;
  private bool _is_quiescent = false;

  /**
   * The type of persona store this is.
   *
   * See {@link Folks.PersonaStore.type_id}.
   */
  public override string type_id { get { return BACKEND_NAME; } }

  /**
   * Whether this PersonaStore can add {@link Folks.Persona}s.
   *
   * See {@link Folks.PersonaStore.can_add_personas}.
   *
   * @since 0.3.1
   */
  public override MaybeBool can_add_personas
  {
    get { return MaybeBool.FALSE; }
  }

  /**
   * Whether this PersonaStore can set the alias of {@link Folks.Persona}s.
   *
   * See {@link Folks.PersonaStore.can_alias_personas}.
   *
   * @since 0.3.1
   */
  public override MaybeBool can_alias_personas
  {
    get { return MaybeBool.FALSE; }
  }

  /**
   * Whether this PersonaStore can set the groups of {@link Folks.Persona}s.
   *
   * See {@link Folks.PersonaStore.can_group_personas}.
   *
   * @since 0.3.1
   */
  public override MaybeBool can_group_personas
  {
    get { return MaybeBool.FALSE; }
  }

  /**
   * Whether this PersonaStore can remove {@link Folks.Persona}s.
   *
   * See {@link Folks.PersonaStore.can_remove_personas}.
   *
   * @since 0.3.1
   */
  public override MaybeBool can_remove_personas
  {
    get { return MaybeBool.FALSE; }
  }

  /**
   * Whether this PersonaStore has been prepared.
   *
   * See {@link Folks.PersonaStore.is_prepared}.
   *
   * @since 0.3.0
   */
  public override bool is_prepared
  {
    get { return this._is_prepared; }
  }

  /**
   * {@inheritDoc}
   *
   * @since 0.6.2
   */
  public override string[] always_writeable_properties
  {
    get { return this._always_writeable_properties; }
  }

  /*
   * Whether this PersonaStore has reached a quiescent state.
   *
   * See {@link Folks.PersonaStore.is_quiescent}.
   *
   */
  public override bool is_quiescent
  {
    get { return this._is_quiescent; }
  }

  /**
   * The {@link Persona}s exposed by this PersonaStore.
   *
   * See {@link Folks.PersonaStore.personas}.
   */
  public override Map<string, Folks.Persona> personas
  {
    get { return this._personas_ro; }
  }

  /**
   * Create a new PersonaStore.
   *
   * Create a new persona store to store the {@link Persona}s for the contacts
   * in purple account provided by ``account``.
   *
   * @param account the purple account being represented by the persona store
   */
  public PersonaStore (int account, PurpleInterface purple) throws GLib.Error
  {
    Object (display_name: purple.purple_account_get_name_for_display (account),
    id: account.to_string(), protocol_name: purple.purple_account_get_protocol_name (account));

    this._purple = purple;
    this.account = account;
  }

  construct
  {
    debug ("Creating new Purple.PersonaStore %p ('%s') for PurpleAccount %p.",
    this, this.id, this.display_name);
    this._personas = new HashMap<string, Persona> ();
    this._personas_ro = this._personas.read_only_view;
    this._buddy_persona_map = new HashMap<int, Persona> ();
  }

  ~PersonaStore ()
  {
    debug ("Destroying Purple.PersonaStore %p ('%s').", this, this.id);
  }

  /**
   * Prepare the PersonaStore for use.
   *
   * See {@link Folks.PersonaStore.prepare}.
   *
   * @throws GLib.Error currently unused
   */
  public override async void prepare () throws GLib.Error
  {
    if (this._is_prepared || this._prepare_pending)
    {
      return;
    }

    try
    {
      this._prepare_pending = true;
      var added_personas = new HashSet<Persona> ();

      // Load the purple buddies
      var blist = this._purple.purple_blist_get_buddies ();

      foreach (var buddy in blist) {
        var p = new Persona (buddy, this, this._purple);
        this._personas.set (p.iid, p);
        this._buddy_persona_map.set (buddy, p);
        added_personas.add (p);
      }

      this._emit_personas_changed (added_personas, null);

      this._is_prepared = true;
      this.notify_property ("is-prepared");

      this._is_quiescent = true;
      this.notify_property ("is-quiescent");
    }
    finally
    {
      this._prepare_pending = false;
    }
  }

  internal void _add_persona (int buddy) {
    var persona = this._buddy_persona_map.get (buddy);
    if (persona != null)
      return;
    try
    {
      var added_personas = new SmallSet<Persona> ();
      var p = new Persona (buddy, this, this._purple);
      this._personas.set (p.iid, p);
      this._buddy_persona_map.set (buddy, p);
      added_personas.add (p);
      this._emit_personas_changed (added_personas, null);
    } catch (Error e) {
      error ("Error couldn't add buddy: %s", e.message);
    }
  }

  internal void _remove_persona (int buddy) {
    var p = this._buddy_persona_map.get (buddy);
    if (p == null)
      return;

    var removed_personas = new SmallSet<Persona> ();
    this._personas.unset (p.iid);
    this._buddy_persona_map.unset (buddy);
    removed_personas.add (p);
    this._emit_personas_changed (null, removed_personas);
  }

  /**
   * Remove a {@link Persona} from the PersonaStore.
   *
   * See {@link Folks.PersonaStore.remove_persona}.
   *
   * @throws Folks.PersonaStoreError.UNSUPPORTED_ON_USER if ``persona`` is the
   * local user — removing the local user isn’t supported
   * @throws Folks.PersonaStoreError.REMOVE_FAILED if removing the contact
   * failed
   */
  public override async void remove_persona (Folks.Persona persona)
    throws Folks.PersonaStoreError
    {
      throw new PersonaStoreError.READ_ONLY (
        "Personas cannot be removed from this store.");
    }

 /**
   * Add a new {@link Persona} to the PersonaStore.
   *
   * See {@link Folks.PersonaStore.add_persona_from_details}.
   *
   * @param details a map of keys to values giving the persona’s initial details
   * @throws Folks.PersonaStoreError.READ_ONLY every time since the
   * purple backend is read-only.
   *
   */
  public override async Folks.Persona? add_persona_from_details (
    HashTable<string, Value?> details) throws Folks.PersonaStoreError
  {
    throw new PersonaStoreError.READ_ONLY (
      "Personas cannot be added to this store.");

  }
}
