/*
 * Copyright (C) 2019 Purism SPC
 * SPDX-License-Identifier: LGPL-2.1
 * Author: Julian Sparber <julian.sparber@puri.sm>
 * 
 * This file was copied from the telepathy backend and then modified
 */

using Gee;
using GLib;
using Folks;

/**
 * A persona subclass which represents a single buddy from purple.
 *
 * There is a one-to-one correspondence between {@link Purple.Persona}s and
 * {@link Purple.Buddy}s
 */
public class Folks.Backends.Purple.Persona : Folks.Persona,
    AliasDetails,
    AvatarDetails,
    ImDetails
{
  private PurpleInterface _purple = null;

  private int buddy { get; set; }
  private const string[] _linkable_properties =
    {
      "im-addresses",
      null /* FIXME: https://bugzilla.gnome.org/show_bug.cgi?id=682698 */
    };
  private string[] _writeable_properties = null;

  private LoadableIcon? _avatar = null;

  /**
   * An avatar for the Persona.
   *
   * See {@link Folks.AvatarDetails.avatar}.
   *
   * @since 0.6.0
   */
  [CCode (notify = false)]
  public LoadableIcon? avatar
    {
      get { return this._avatar; }
      set { this.change_avatar.begin (value); } /* not writeable */
    }

  /**
   * The Persona's presence type.
   *
   * See {@link Folks.PresenceDetails.presence_type}.
   */
  public Folks.PresenceType presence_type { get; set; }

  /**
   * The Persona's presence status.
   *
   * See {@link Folks.PresenceDetails.presence_status}.
   *
   * @since 0.6.0
   */
  public string presence_status { get; set; }

  /**
   * The Persona's presence message.
   *
   * See {@link Folks.PresenceDetails.presence_message}.
   */
  public string presence_message { get; set; }

  /**
   * The Persona's client types.
   *
   * See {@link Folks.PresenceDetails.client_types}.
   *
   * @since 0.9.5
   */
  public string[] client_types { get; set; }

  /**
   * The names of the Persona's linkable properties.
   *
   * See {@link Folks.Persona.linkable_properties}.
   */
  public override string[] linkable_properties
    {
      get { return Purple.Persona._linkable_properties; }
    }

  /**
   * {@inheritDoc}
   *
   * @since 0.6.0
   */
  public override string[] writeable_properties
    {
      get { return this._writeable_properties; }
    }

  private string _alias = ""; /* must never be null */

  /**
   * An alias for the Persona.
   *
   * See {@link Folks.AliasDetails.alias}.
   */
  [CCode (notify = false)]
  public string alias
    {
      get { return this._alias; }
      set { this.change_alias.begin (value); }
    }

  /* NOTE: Other properties support lazy initialisation, but im-addresses
   * doesn't as it's a linkable property, so always has to be loaded anyway. */
  private HashMultiMap<string, ImFieldDetails> _im_addresses =
      new HashMultiMap<string, ImFieldDetails> (null, null,
          AbstractFieldDetails<string>.hash_static,
          AbstractFieldDetails<string>.equal_static);

  /**
   * A mapping of IM protocol to an (unordered) set of IM addresses.
   *
   * See {@link Folks.ImDetails.im_addresses}.
   */
  [CCode (notify = false)]
  public MultiMap<string, ImFieldDetails> im_addresses
    {
      get { return this._im_addresses; }
      set { this.change_im_addresses.begin (value); }
    }

  private uint _im_interaction_count = 0;

  /**
   * A counter for IM interactions (send/receive message) with the persona.
   *
   * See {@link Folks.InteractionDetails.im_interaction_count}
   *
   * @since 0.7.1
   */
  public uint im_interaction_count
    {
      get { return this._im_interaction_count; }
    }

  internal DateTime? _last_im_interaction_datetime = null;

  /**
   * The latest datetime for IM interactions (send/receive message) with the
   * persona.
   *
   * See {@link Folks.InteractionDetails.last_im_interaction_datetime}
   *
   * @since 0.7.1
   */
  public DateTime? last_im_interaction_datetime
    {
      get { return this._last_im_interaction_datetime; }
    }

  private uint _call_interaction_count = 0;

  /**
   * A counter for call interactions (only successful calls) with the persona.
   *
   * See {@link Folks.InteractionDetails.call_interaction_count}
   *
   * @since 0.7.1
   */
  public uint call_interaction_count
    {
      get { return this._call_interaction_count; }
    }

  internal DateTime? _last_call_interaction_datetime = null;

  /**
   * The latest datetime for call interactions (only successful calls) with the
   * persona.
   *
   * See {@link Folks.InteractionDetails.last_call_interaction_datetime}
   *
   * @since 0.7.1
   */
  public DateTime? last_call_interaction_datetime
    {
      get { return this._last_call_interaction_datetime; }
    }

  /**
   * Create a new persona.
   *
   * Create a new persona for the {@link PersonaStore} ``store``, representing
   * the Telepathy contact given by ``contact``.
   *
   * @param contact the Telepathy contact being represented by the persona
   * @param store the persona store to place the persona in
   */
  public Persona (int buddy, PersonaStore store, PurpleInterface purple) throws GLib.Error
  {
    string id = purple.purple_buddy_get_name (buddy);
    var uid = Folks.Persona.build_uid (store.type_id, store.id, id);
    var is_user = false;

    Object (display_id: id,
            /* FIXME: This IID format should be moved out to the ImDetails
             * interface along with the code in
             * Kf.Persona.linkable_property_to_links(), but that depends on
             * bgo#624842 being fixed. */
            iid: store.protocol_name + ":" + id,
            uid: uid,
            store: store,
            is_user: is_user);

    debug ("Created new Purple.Persona '%s' for service-specific UID '%s': %p",
           uid, id, this);

    this.buddy = buddy;
    this._purple = purple;
    _build_buddy ();
  }

  ~Persona ()
  {
    debug ("Destroying Purple.Persona '%s': %p", this.uid, this);
  }

  private void _build_buddy () {
    try {
      File icon_file = null;
      this._alias = this._purple.purple_buddy_get_alias (this.buddy);

      int icon = this._purple.purple_buddy_get_icon (this.buddy);
      if (icon != 0) {
        string icon_path = this._purple.purple_buddy_icon_get_full_path (icon);
        icon_file = File.new_for_path (icon_path);
      }
      if (icon_file != null)
        this._avatar = new FileIcon (icon_file);

      debug ("All details are loaded");
    } catch (GLib.Error e) {
      error ("Error loading infromation for purple contact: %s", e.message);
    }

    _build_im_address ();
  }

  private void _build_im_address () {
    try {
      var value = this._purple.purple_buddy_get_name (this.buddy);
      var address = new ImFieldDetails (value);
      var purple_store = this.store as Purple.PersonaStore;
      if (purple_store != null)
        this.im_addresses.set (purple_store.protocol_name, address);
    } catch (GLib.Error e) {
      error ("Error loading infromation for purple contact: %s", e.message);
    }
  }
}
